# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from pai4sk.cluster.k_means_ import KMeansSklearn
from ...exceptions import NotFittedError
import warnings
import numbers
import numpy as np
from .utils import _can_allow_cuml, _is_cuml_algorithm, _do_not_allow_cuml

class KMeans(KMeansSklearn):
    """K-Means clustering.

    If cuml is installed and cudf dataframe is passed as input, then pai4sk
    will try to use the accelerated KMeans algorithm from cuML. Otherwise,
    scikit-learn's KMeans algorithm will be used.

    cuML in pai4sk is currently supported only
    |  (a) with python 3.6 and
    |  (b) without MPI.
    |  If KMeans from cuML is run, then the return values from the APIs will be
    cudf dataframe and cudf Series objects instead of the return types of
    scikit-learn API.

    Parameters
    ----------

    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of
        centroids to generate.

    init : {'k-means++', 'random' or an ndarray}
        Method for initialization, defaults to 'k-means++':

        'k-means++' : selects initial cluster centers for k-mean
        clustering in a smart way to speed up convergence. See section
        Notes in k_init for more details.

        'random': choose k observations (rows) at random from data for
        the initial centroids.

        If an ndarray is passed, it should be of shape (n_clusters, n_features)
        and gives the initial centers.

    n_init : int, default: 10
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    max_iter : int, default: 300
        Maximum number of iterations of the k-means algorithm for a
        single run.

    tol : float, default: 1e-4
        Relative tolerance with regards to inertia to declare convergence

    precompute_distances : {'auto', True, False}
        Precompute distances (faster but takes more memory).

        'auto' : do not precompute distances if n_samples * n_clusters > 12
        million. This corresponds to about 100MB overhead per job using
        double precision.

        True : always precompute distances

        False : never precompute distances

    verbose : int, default 0
        Verbosity mode.

    random_state : int, RandomState instance or None (default)
        Determines random number generation for centroid initialization. Use
        an int to make the randomness deterministic.
        See :term:`Glossary <random_state>`.

    copy_x : boolean, optional
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True (default), then the original data is
        not modified, ensuring X is C-contiguous.  If False, the original data
        is modified, and put back before the function returns, but small
        numerical differences may be introduced by subtracting and then adding
        the data mean, in this case it will also not ensure that data is
        C-contiguous which may cause a significant slowdown.

    n_jobs : int or None, optional (default=None)
        The number of jobs to use for the computation. This works by computing
        each of the n_init runs in parallel.

        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    algorithm : "auto", "full" or "elkan", "cuml", default="auto"
        K-means algorithm to use. The classical EM-style algorithm is "full".
        The "elkan" variation is more efficient by using the triangle
        inequality, but currently doesn't support sparse data. "auto" chooses
        "elkan" for dense data and "full" for sparse data.

        If cuml is installed and cudf dataframe is passed as input, then if either
        |  (1) algorithm is set to "cuml" or
        |  (2) algorithm is "auto",
        |  then pai4sk will try to use kmeans algorithm from RAPIDS cuML.
        cuML in pai4sk is currently supported only
        |  (a) with python 3.6 and
        |  (b) without MPI.
        |  If KMeans from cuML is run, then the return values of the APIs will be
        cudf dataframe and cudf Series objects instead of the return types of
        scikit-learn API.

    Attributes
    ----------
    cluster_centers_ : array, [n_clusters, n_features] or cudf dataframe
        Coordinates of cluster centers. If the algorithm stops before fully
        converging (see ``tol`` and ``max_iter``), these will not be
        consistent with ``labels_``. If KMeans from cuML is run, then the
        return values of some of the APIs will be cudf dataframe and
        cudf Series objects instead of the return types of scikit-learn API.

    labels_ : array or cudf Series
        Labels of each point

    inertia_ : float
        Sum of squared distances of samples to their closest cluster center.

    n_iter_ : int
        Number of iterations run.

    use_gpu : boolean, Default is True
        If True, cuML will use all GPUs. Applicable only for cuML.

    Examples
    --------

    >>> from pai4sk.cluster import KMeans
    >>> import numpy as np
    >>> X = np.array([[1, 2], [1, 4], [1, 0],
    ...               [4, 2], [4, 4], [4, 0]])
    >>> kmeans = KMeans(n_clusters=2, random_state=0).fit(X)
    >>> kmeans.labels_
    array([0, 0, 0, 1, 1, 1], dtype=int32)
    >>> kmeans.predict([[0, 0], [4, 4]])
    array([0, 1], dtype=int32)
    >>> kmeans.cluster_centers_
    array([[1., 2.],
           [4., 2.]])

    See also
    --------

    MiniBatchKMeans
        Alternative online implementation that does incremental updates
        of the centers positions using mini-batches.
        For large scale learning (say n_samples > 10k) MiniBatchKMeans is
        probably much faster than the default batch implementation.

    Notes
    ------
    The k-means problem is solved using either Lloyd's or Elkan's algorithm.

    The average complexity is given by O(k n T), were n is the number of
    samples and T is the number of iteration.

    The worst case complexity is given by O(n^(k+2/p)) with
    n = n_samples, p = n_features. (D. Arthur and S. Vassilvitskii,
    'How slow is the k-means method?' SoCG2006)

    In practice, the k-means algorithm is very fast (one of the fastest
    clustering algorithms available), but it falls in local minima. That's why
    it can be useful to restart it several times.

    If the algorithm stops before fully converging (because of ``tol`` or
    ``max_iter``), ``labels_`` and ``cluster_centers_`` will not be consistent,
    i.e. the ``cluster_centers_`` will not be the means of the points in each
    cluster. Also, the estimator will reassign ``labels_`` after the last
    iteration to make ``labels_`` consistent with ``predict`` on the training
    set.

    """

    __slots__ = ('_KMeansSklearn')

    def __init__(self, n_clusters=8, max_iter=300, tol=1e-4, verbose=0, random_state=1, precompute_distances='auto', init='k-means++', n_init=1, algorithm='auto', copy_x=True, n_jobs=None, use_gpu=True):

        self.n_clusters = n_clusters
        self.verbose = verbose
        self.random_state = random_state
        self.precompute_distances = precompute_distances
        self.init = init
        self.n_init = n_init
        self.copy_x = copy_x
        self.n_jobs = n_jobs
        self.algorithm = algorithm
        self.max_iter = max_iter
        self.tol = tol
        self.use_gpu = use_gpu
        self._cumlKM = None
        self.cluster_centers_ = None
        self.labels_ = None
        self.inertia_ = 0.0
        self.n_iter = 0
        self._KMeansSklearn = super(KMeans, self)
        # Perform basic validations for enabling cuML first
        self._cuml_basic_validations = _can_allow_cuml()

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._KMeansSklearn = super(KMeans, self)

    def fit(self, X, y=None, sample_weight=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features) or cudf dataframe
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like, shape (n_samples,) or (n_samples, n_targets)
            Target vector relative to X.
        sample_weight : array-like, shape (n_samples,) optional
            Array of weights that are assigned to individual samples.
            If not provided, then each sample is given unit weight.
            .. versionadded:: 0.17
               *sample_weight* support to KMeans.
        Returns
        -------
        self : object
            If KMeans from cuML is run then this fit method saves the cluster
            centers and labels as cudf dataframe and cudf Series objects
            instead of the return types of scikit-learn API.
        """

        self._validate_fit_args()

        if self._check_set_cuml(X, sample_weight):
            print("Info: Invoking cuML kmeans fit...")
            self._cumlKM.fit(X)
            self._save_cuml_fit_results()
        else:
            _do_not_allow_cuml(X)
            print("Info: Invoking scikit-learn kmeans fit...")
            self._KMeansSklearn.fit(X, y, sample_weight)

        return self

    def predict(self, X, sample_weight=None):
        """
        Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, cluster_centers_ is called
        the code book and each value returned by predict is the index of
        the closest code in the code book.
        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            cuDF dataframe if cuml is installed.
            New data to predict.
        sample_weight : array-like, shape (n_samples,), optional
                        The weights for each observation in X. If None, all
                        observations are assigned equal weight (default: None)

        Returns:
        labels : array, shape [n_samples,] or cudf Series object
                 Index of the cluster each sample belongs to.
                 If KMeans from cuML is run, then this method returns
                 cudf Series object instead of the return types of
                 scikit-learn API.
        """

        # cluster_centers_ is computed in fit()
        if self.cluster_centers_ is None:
            raise NotFittedError("Call fit before prediction")

        if self._cumlKM:
            return self._cumlKM.predict(X)
        else:
            return self._KMeansSklearn.predict(X, sample_weight)

    def fit_predict(self, X, y=None, sample_weight=None):
        """
        Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by predict(X).
        Parameters:

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            cuDF dataframe if cuml is installed.
            New data to transform.
        y : Ignored
            not used, present here for API consistency by convention.
        sample_weight : array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all
            observations are assigned equal weight (default: None)

        Returns:
        labels : array, shape [n_samples,] or cudf Series object
                 Index of the cluster each sample belongs to.
                 If KMeans from cuML is run, then this method saves the cluster
                 centers and labels as cudf dataframe and cudf Series objects
                 instead of the return types of scikit-learn API. Returns cudf
                 Series object.
        """

        self._validate_fit_args()

        if self._check_set_cuml(X, sample_weight):
            print("Info: Invoking cuML kmeans fit_predict...")
            self._cumlKM.fit_predict(X)
            self._save_cuml_fit_results()
            return self.labels_
        else:
            _do_not_allow_cuml(X)
            print("Info: Invoking scikit-learn kmeans fit_predict...")
            return self._KMeansSklearn.fit_predict(X)

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            cuDF dataframe if cuml is installed.
            New data to transform.
            If KMeans from cuML is run and if the input data is a cudf dataframe,
            then this method returns cudf dataframe instead of array.
        Returns
        -------
        X_new : array, shape [n_samples, k] or cudf dataframe
                X transformed in the new space.
                If KMeans from cuML is run, then this method returns cudf
                dataframe instead of the return types of scikit-learn API.
        """
        if self._cumlKM:
            return self._cumlKM.transform(X)
        else:
            return self._KMeansSklearn.transform(X)

    def fit_transform(self, X, y=None, sample_weight=None):
        """
        Compute clustering and transform X to cluster-distance space.

        Equivalent to fit(X).transform(X), but more efficiently implemented.
        Parameters:

        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            cuDF dataframe if cuml is installed.
            New data to transform.
        y : Ignored
            not used, present here for API consistency by convention.
        sample_weight : array-like, shape (n_samples,), optional
            The weights for each observation in X. If None, all
            observations are assigned equal weight (default: None)

        Returns:
        X_new : array, shape [n_samples, k] or cudf dataframe
               X transformed in the new space.
               If KMeans from cuML is run, then this method saves the cluster
               centers and labels as cudf dataframe and cudf Series objects
               instead of the return types of scikit-learn API.
        """

        self._validate_fit_args()

        if self._check_set_cuml(X, sample_weight):
            print("Info: Invoking cuML kmeans fit_transform...")
            X_new = self._cumlKM.fit_transform(X)
            self._save_cuml_fit_results()
            return X_new
        else:
            _do_not_allow_cuml(X)
            print("Info: Invoking scikit-learn kmeans fit_transform...")
            return self._KMeansSklearn.fit_transform(X)

    def _check_set_cuml(self, X, sample_weight=None):
        """
           Check if cuML can be used for this k means
        """

        # Check if basic validations for enabling cuML succeeded
        if not self._cuml_basic_validations:
            return False

        # Perform further vaildations for enabling cuML
        import cudf
        if not isinstance(X, cudf.DataFrame):
            warnings.warn("Currently cuML in pai4sk can work only on " \
                          "cudf.DataFrame.", UserWarning)
            return False

        if not self.use_gpu:
            warnings.warn("Conflicting arguments are seen. cudf.DataFrame" \
                          " is given as input and use_gpu is set to False.",
                          UserWarning)
            return False

        if not _is_cuml_algorithm(self.algorithm):
            return False

        if sample_weight is not None:
            warnings.warn("Currently cuML in pai4sk can't handle " \
                          "sample_weight argument.", UserWarning)
            return False

        if (self.init != 'kmeans++' and self.init != 'k-means++'):
            warnings.warn("Currently cuML in pai4sk can't handle 'init' " \
                          "parameter with values other than 'k-means++' " \
                          "and 'kmeans++'.", UserWarning)
            return False

        if not self._cumlKM:
            from pai4sk import sml_utils
            ndevices = sml_utils.get_gpu_count()
            if ndevices == 0:
                warnings.warn("Currently cuML in pai4sk can work only with GPUs." \
                              "No GPUs are available.", UserWarning)
                return False

            if (self.init == 'k-means++'):
                self.init = 'kmeans++'
            from cuml import KMeans as cumlkm
            n_gpu = -1
            self._cumlKM = cumlkm(self.n_clusters, self.max_iter, self.tol, self.verbose,
                                 self.random_state, self.precompute_distances, self.init,
                                 self.n_init, self.algorithm, n_gpu)
            warnings.warn("Using KMeans from cuML.", UserWarning)

        return True

    def _validate_fit_args(self):
        """Validate the parameters used by fit method
        """
        if not isinstance(self.max_iter, numbers.Number) or self.max_iter < 0:
            raise ValueError('Number of iterations should be a positive number,'
                         ' got %d instead' % self.max_iter)

        if not isinstance(self.tol, numbers.Number) or self.tol < 0:
            raise ValueError("Tolerance for stopping criteria must be "
                             "positive; got (tol=%r)" % self.tol)

        if self.n_init <= 0:
            raise ValueError("Invalid number of initializations."
                             " n_init=%d must be bigger than zero." % self.n_init)

        if self.precompute_distances != 'auto' and not isinstance(self.precompute_distances, bool):
            raise ValueError("precompute_distances should be 'auto' or True/False, "
                             "but a value of %r was passed" % self.precompute_distances)

    def _save_cuml_fit_results(self):
        """Save cluster centers and labels"""

        self.cluster_centers_ = self._cumlKM.cluster_centers_
        self.labels_ = self._cumlKM.labels_

