# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

# Sample program to use DBSCAN from cuML in pai4sk

import sys
if sys.version_info.major != 3:
    raise ValueError("cuML in pai4sk is supported only with python 3.6.")

import numpy as np
import pandas as pd
from pai4sk.cluster import DBSCAN as pai4skDBSCAN
from sklearn.cluster import DBSCAN as skDBSCAN
import cudf
import os
import time
import argparse

CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--eps",
   type=float,
   default=3.65
)
CLI.add_argument(
   "--nrows",
   type=int,
   default=10000
)
CLI.add_argument(
   "--ncols",
   type=int,
   default=128
)
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
args = CLI.parse_args()

import gzip
def load_data(nrows, ncols, cached = 'data/mortgage.npy.gz'):
    if os.path.exists(cached):
        print('use mortgage data')
        with gzip.open(cached) as f:
            X = np.load(f)
        X = X[np.random.randint(0,X.shape[0]-1,nrows),:ncols]
    else:
        print('use random data')
        X = np.random.rand(nrows,ncols)
    df = pd.DataFrame({'fea%d'%i:X[:,i] for i in range(X.shape[1])})
    return df

from sklearn.metrics import mean_squared_error
def array_equal(a,b,threshold=5e-3,with_sign=True):
    a = to_nparray(a)
    b = to_nparray(b)
    if with_sign == False:
        a,b = np.abs(a),np.abs(b)
    res = mean_squared_error(a,b)<threshold
    return res

def to_nparray(x):
    if isinstance(x,np.ndarray) or isinstance(x,pd.DataFrame):
        return np.array(x)
    elif isinstance(x,np.float64):
        return np.array([x])
    elif isinstance(x,cudf.DataFrame) or isinstance(x,cudf.Series):
        return x.to_pandas().values
    return x

nrows = args.nrows
ncols = args.ncols

X1 = load_data(nrows,ncols)
print('data',X1.shape)

eps = args.eps
min_samples = 2

X = cudf.DataFrame.from_pandas(X1)

clustering_pai4sk = pai4skDBSCAN(eps = eps, min_samples = min_samples)
t0 = time.time()
clustering_pai4sk.fit(X)
print("[pai4sk] Training time (s):  {0:.2f}".format(time.time()-t0))

if args.skip_sklearn:
    print('Execution using scikit-learn is skipped.')
    exit()

clustering_sk = skDBSCAN(eps = eps, min_samples = min_samples)
t0 = time.time()
clustering_sk.fit(X1)
print("[sklearn] Training time (s):  {0:.2f}".format(time.time()-t0))

passed = array_equal(clustering_sk.labels_,clustering_pai4sk.labels_)
message = 'compare dbscan: pai4sk vs sklearn labels %s'%('equal'if passed else 'NOT equal')
print(message)
