# Criteo-Kaggle Example

In this example we will train a Logistic Regression on the criteo-kaggle dataset, using both `scikit-learn` (`sklearn`) and `snap-ml-mpi`. 
Note that `scikit-learn` does not provide an MPI implementation. We will however still perform this comparison in order to mainly verify the accuracy metrics. If you want to skip the scikit-learn part of the examples, you can use --skip_sklearn option on the command line of examples. This is useful specifically when large data is being used by example and in such cases scikit-learn can take a long time to perform training.

The criteo-kaggle dataset is from the [Display Advertising Challenge](https://www.kaggle.com/c/criteo-display-ad-challenge). 

The contest is described as follows:

> Display advertising is a billion dollar effort and one of the central uses of machine learning on the Internet. However, its data and methods are usually kept under lock and key. In this research competition, CriteoLabs is sharing a week’s worth of data for you to develop models predicting ad click-through rate (CTR). Given a user and the page he is visiting, what is the probability that he will click on a given ad?


## Download the Data

Download and decompress the data:
```bash
mkdir data
cd data
wget https://s3-us-west-2.amazonaws.com/criteo-public-svm-data/criteo.kaggle2014.svm.tar.gz
tar xzf criteo.kaggle2014.svm.tar.gz
cd ../
```
Now one should see a file named `criteo.kaggle2014.svm` in the `data/` directory.

## Data Pre-processing

First we split the data into train and test datasets.
Next we write the train/test data to disk as binary scipy/numpy files. This helps to achieve fast data loading. 
This format will be used for loading the datasets using `sklearn`. `sklearn` will be used for performance comparison.
Finally we write the train and test datasets in distributed snap format which will be used only for training the `pai4sk` Logistic Regression model. 
This format helps achieve higher performance.

The entire pre-processing code can be run by executing `snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-criteo/preprocess-criteo.py`

A summary of all the pre-processed files is shown below:
```bash
ls -l data/
criteo.kaggle2014.test.snap
criteo.kaggle2014.train.snap  
criteo.kaggle2014.train.svm  
criteo.kaggle2014.X_test.npz  
criteo.kaggle2014.X_train.npz  
criteo.kaggle2014.y_test.npy
criteo.kaggle2014.y_train.npy
```

## Load pre-processed data

After running the data pre-processing steps above, we are ready to perform our benchmarking: `snap.ml` vs. `sklearn`.
Pre-processed data can be loaded with below methods. See API documentation of the following methods for more details.
```
1.`load_from_snap_format`
See example-criteo.py
2.`load_snap_file`
See example-criteo-snap-file.py
3.`load_from_svmlight_format`
See example-criteo-svmlight-format.py
4.`load_svmlight_file`
See example-criteo-svmlight-file.py
```

`load_from_snap_format` and `load_from_svmlight_format` return data_ptr.  data_ptr will be of type DensePartition, SparsePartition or ConstantValueSparsePartition.

`load_snap_file` returns dense or sparse matrices as X,y. Here X will be npy, csr or csc and y will be npy.

`load_svmlight_file` returns sparse matrix X and numpy array y. Here X will be csr or csc.

Note: dense snap data can be loaded with `load_from_snap_format` and `load_snap_file` methods.

## Training and Evaluating a Logistic Regression Model

After running the data pre-processing steps above, we are ready to perform our benchmarking: `snap.ml` vs. `sklearn`.

First we train a Logistic Regression model using the IBM `pai4sk` library. 
We provide command line option --use_gpu to run the example for how to train the model using 1) GPUs or 2) CPUs.

We define the data paths for the train and test datasets in distributed snap format.

You need to have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).

Next we read the datasets from snap format using the loading functions provided by the `pai4sk` library and we measure the data loading time.

After initialization, we call the `fit()` function in order to train the dual Logistic Regression model. We also time the execution of the training process.

Please note that the `pai4sk` library assumes for maximum performance to run 1 MPI process per node.

As a last step we make predictions on the test data and we compute the accuracy and the logistic loss metrics. 
The `pai4sk` library provides distributed implementations of the accuracy and logistic loss metrics.

Note that the `pai4sk` CPU algorithm converges significantly faster than the
`pai4sk` GPU algorithm with respect to the number of epochs (iterations). Therefore
it is sufficient to set the `max_iter` parameter to 6.
Moreover, the CPU algorithm supports multi-threading training. To set the number of threads change the `num_threads` variable. 

If the CPU training does not converge, set the `num_threads` to 1 or use the GPU algorithm.

Then we train a Logistic Regression model using the `sklearn` library.

## Results

To run on one machine with 1 MPI process, one should run the following command: 
```bash
snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-criteo/example-criteo.py
```

To run on 2 machines with 1 MPI process per machine, one should run the following command:
```bash
snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-criteo/example-criteo.py
```

The user can also run multiple MPI processes per machine, but, in that case, the MPI processes on one machine will share the same hardware resources during training. Namely, if the user trains the model on GPUs, the MPI processes on one machine will share the same GPUs as defined in the `device_ids` parameter. Therefore, we recommend the user to run 1 MPI process per machine in order to achieve the maximum performance.


One should see the following results on a **P9+V100** system when running `snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-criteo/example-criteo.py`:

```
[pai4sk] [Data loading] Time = 3.9546
[pai4sk] [GPU Training] Accuracy = 0.7877
[pai4sk] [GPU Training] LogLoss = 0.4558
[pai4sk] [GPU Training] Training time = 23.3201

[sklearn] [Data loading] Time = 30.3557
[sklearn] Accuracy = 0.7877
[sklearn] LogLoss = 0.4558
[sklearn] Training time = 558.98
```

One should see the following results on 2 **P9+V100** systems when running `snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-criteo/example-criteo.py`:

```
[pai4sk] [GPU Training] Accuracy = 0.7877
[pai4sk] [GPU Training] LogLoss = 0.4558
[pai4sk] [GPU Training] Training time = 21.9170

[sklearn] [Data loading] Time = 30.3557
[sklearn] Accuracy = 0.7877
[sklearn] LogLoss = 0.4558
[sklearn] Training time = 558.98
```

Please note that the reported times can vary depending on the system architecture.

&copy; Copyright IBM Corporation 2018, 2019
